@@ -5,10 +5,11 @@ use strict;
 use warnings; no warnings qw(void once uninitialized numeric redefine);
 
 our $AUTHORITY = 'cpan:TOBYINK';
-our $VERSION   = '0.036';
-our @EXPORT_OK = qw< mkopt mkopt_hash _croak >;
+our $VERSION   = '0.040';
+our @EXPORT_OK = qw< mkopt mkopt_hash _croak _carp >;
 
 sub _croak ($;@) { require Carp; my $fmt = shift; @_ = sprintf($fmt, @_); goto \&Carp::croak }
+sub _carp  ($;@) { require Carp; my $fmt = shift; @_ = sprintf($fmt, @_); goto \&Carp::carp }
 
 sub import
 {
@@ -17,17 +18,29 @@ sub import
 	my @args        = do { no strict qw(refs); @_ ? @_ : @{"$class\::EXPORT"} };
 	my $opts        = mkopt(\@args);
 	
-	$global_opts->{into} = caller unless exists $global_opts->{into};
 	my @want;
+	my %not_want;
+	$global_opts->{into} = caller unless exists $global_opts->{into};
+	$global_opts->{not}  = \%not_want;
 	
 	while (@$opts)
 	{
 		my $opt = shift @{$opts};
 		my ($name, $value) = @$opt;
 		
-		$name =~ /^[:-](.+)$/
-			? push(@$opts, $class->_exporter_expand_tag($1, $value, $global_opts))
-			: push(@want, $opt);
+		($name =~ m{\A\!(/.+/[msixpodual]+)\z}) ?
+			do {
+				my @not = $class->_exporter_expand_regexp($1, $value, $global_opts);
+				++$not_want{$_->[0]} for @not;
+			} :
+		($name =~ m{\A\!(.+)\z}) ?
+			(++$not_want{$1}) :
+		($name =~ m{\A[:-](.+)\z}) ?
+			push(@$opts, $class->_exporter_expand_tag($1, $value, $global_opts)) :
+		($name =~ m{\A/.+/[msixpodual]+\z}) ?
+			push(@$opts, $class->_exporter_expand_regexp($name, $value, $global_opts)) :
+		# else ?
+			push(@want, $opt);
 	}
 	
 	$class->_exporter_validate_opts($global_opts);
@@ -35,6 +48,8 @@ sub import
 	
 	for my $wanted (@want)
 	{
+		next if $not_want{$wanted->[0]};
+		
 		my %symbols = $class->_exporter_expand_sub(@$wanted, $global_opts, $permitted);
 		$class->_exporter_install_sub($_, $wanted->[1], $global_opts, $symbols{$_})
 			for keys %symbols;
@@ -50,6 +65,32 @@ sub _exporter_validate_opts
 	1;
 }
 
+# Called after expanding a tag or regexp to merge the tag's options with
+# any sub-specific options.
+sub _exporter_merge_opts
+{
+	my $class = shift;
+	my ($tag_opts, $global_opts, @stuff) = @_;
+	
+	$tag_opts = {} unless ref($tag_opts) eq q(HASH);
+	_croak('Cannot provide an -as option for tags')
+		if exists $tag_opts->{-as};
+	
+	my $optlist = mkopt(\@stuff);
+	for my $export (@$optlist)
+	{
+		next if defined($export->[1]) && ref($export->[1]) ne q(HASH);
+		
+		my %sub_opts = ( %{ $export->[1] or {} }, %$tag_opts );
+		$sub_opts{-prefix} = sprintf('%s%s', $tag_opts->{-prefix}, $export->[1]{-prefix})
+			if exists($export->[1]{-prefix}) && exists($tag_opts->{-prefix});
+		$sub_opts{-suffix} = sprintf('%s%s', $export->[1]{-suffix}, $tag_opts->{-suffix})
+			if exists($export->[1]{-suffix}) && exists($tag_opts->{-suffix});
+		$export->[1] = \%sub_opts;
+	}
+	return @$optlist;
+}
+
 # Given a tag name, looks it up in %EXPORT_TAGS and returns the list of
 # associated functions. The default implementation magically handles tags
 # "all" and "default". The default implementation interprets any undefined
@@ -63,22 +104,36 @@ sub _exporter_expand_tag
 	my ($name, $value, $globals) = @_;
 	my $tags  = \%{"$class\::EXPORT_TAGS"};
 	
-	return map [$_ => $value], $tags->{$name}->($class, @_)
+	return $class->_exporter_merge_opts($value, $globals, $tags->{$name}->($class, @_))
 		if ref($tags->{$name}) eq q(CODE);
 	
-	return map [$_ => $value], @{$tags->{$name}}
+	return $class->_exporter_merge_opts($value, $globals, @{$tags->{$name}})
 		if exists $tags->{$name};
 	
-	return map [$_ => $value], @{"$class\::EXPORT"}, @{"$class\::EXPORT_OK"}
+	return $class->_exporter_merge_opts($value, $globals, @{"$class\::EXPORT"}, @{"$class\::EXPORT_OK"})
 		if $name eq 'all';
 	
-	return map [$_ => $value], @{"$class\::EXPORT"}
+	return $class->_exporter_merge_opts($value, $globals, @{"$class\::EXPORT"})
 		if $name eq 'default';
 	
 	$globals->{$name} = $value || 1;
 	return;
 }
 
+# Given a regexp-like string, looks it up in @EXPORT_OK and returns the
+# list of matching functions.
+# 
+sub _exporter_expand_regexp
+{
+	no strict qw(refs);
+	
+	my $class = shift;
+	my ($name, $value, $globals) = @_;
+	my $compiled = eval("qr$name");
+	
+	$class->_exporter_merge_opts($value, $globals, grep /$compiled/, @{"$class\::EXPORT_OK"});
+}
+
 # Helper for _exporter_expand_sub. Returns a regexp matching all subs in
 # the exporter package which are available for export.
 #
@@ -122,7 +177,7 @@ sub _exporter_fail
 {
 	my $class = shift;
 	my ($name, $value, $globals) = @_;
-	_croak("Could not find sub '$name' to export in package '$class'");
+	_croak("Could not find sub '%s' exported by %s", $name, $class);
 }
 
 # Actually performs the installation of the sub into the target package. This
@@ -148,16 +203,34 @@ sub _exporter_install_sub
 	return ($$name = $sym)                       if ref($name) eq q(SCALAR);
 	return ($into->{$name} = $sym)               if ref($into) eq q(HASH);
 	
-	require B;
-	for (grep ref, $into->can($name))
+	no strict qw(refs);
+	
+	if (exists &{"$into\::$name"} and \&{"$into\::$name"} != $sym)
 	{
-		my $stash = B::svref_2object($_)->STASH;
-		next unless $stash->can("NAME");
-		$stash->NAME eq $into
-			and _croak("Refusing to overwrite local sub '$name' with export from $class");
+		my ($level) = grep defined, $value->{-replace}, $globals->{replace}, q(0);
+		my $action = {
+			carp     => \&_carp,
+			0        => \&_carp,
+			''       => \&_carp,
+			warn     => \&_carp,
+			nonfatal => \&_carp,
+			croak    => \&_croak,
+			fatal    => \&_croak,
+			die      => \&_croak,
+		}->{$level} || sub {};
+		
+		$action->(
+			$action == \&_croak
+				? "Refusing to overwrite existing sub '%s::%s' with sub '%s' exported by %s"
+				: "Overwriting existing sub '%s::%s' with sub '%s' exported by %s",
+			$into,
+			$name,
+			$_[0],
+			$class,
+		);
 	}
 	
-	no strict qw(refs);
+	no warnings qw(prototype);
 	*{"$into\::$name"} = $sym;
 }
 
@@ -200,6 +273,8 @@ __END__
 
 =encoding utf-8
 
+=for stopwords frobnicate greps regexps
+
 =head1 NAME
 
 Exporter::Tiny - an exporter with the features of Sub::Exporter but only core dependencies
@@ -278,8 +353,14 @@ Sub::Exporter, but where it does it usually uses a fairly similar API.
    # call it "my_frobnicate"
    use MyUtils "frobnicate" => { -prefix => "my_" };
 
+   # can set a prefix for *all* functions imported from MyUtils
+   # by placing the options hashref *first*.
+   use MyUtils { prefix => "my_" }, "frobnicate";
+   # (note the lack of hyphen before `prefix`.)
+
    # call it "frobnicate_util"
    use MyUtils "frobnicate" => { -suffix => "_util" };
+   use MyUtils { suffix => "_util" }, "frobnicate";
 
    # import it twice with two different names
    use MyUtils
@@ -319,6 +400,30 @@ OK, Sub::Exporter doesn't do this...
    
    $funcs{frobnicate}->(...);
 
+=head2 DO NOT WANT!
+
+This imports everything except "frobnicate":
+
+   use MyUtils qw( -all !frobnicate );
+
+Negated imports always "win", so the following will not import
+"frobnicate", no matter how many times you repeat it...
+
+   use MyUtils qw( !frobnicate frobnicate frobnicate frobnicate );
+
+=head2 Importing by regexp
+
+Here's how you could import all functions beginning with an "f":
+
+   use MyUtils qw( /^F/i );
+
+Or import everything except functions beginning with a "z":
+
+   use MyUtils qw( -all !/^Z/i );
+
+Note that regexps are always supplied as I<strings> starting with
+C<< "/" >>, and not as quoted regexp references (C<< qr/.../ >>).
+
 =head1 TIPS AND TRICKS EXPORTING USING EXPORTER::TINY
 
 Simple configuration works the same as L<Exporter>; inherit from this module,
@@ -348,7 +453,10 @@ exporter module a class method called C<< _generate_foo >>.
 
 You can also generate tags:
 
-   my %constants = (FOO => 1, BAR => 2);
+   my %constants;
+   BEGIN {
+      %constants = (FOO => 1, BAR => 2);
+   }
    use constant \%constants;
    
    $EXPORT_TAGS{constants} = sub {
@@ -382,6 +490,11 @@ throwing an exception or printing a warning.
 
 The default implementation does nothing interesting.
 
+=item C<< _exporter_merge_opts($tag_opts, $globals, @exports) >>
+
+Called to merge options which have been provided for a tag into the
+options provided for the exports that the tag expanded to.
+
 =item C<< _exporter_expand_tag($name, $args, $globals) >>
 
 This method is called to expand an import tag (e.g. C<< ":constants" >>).
@@ -397,6 +510,13 @@ expansion loop!
 The default implementation uses C<< %EXPORT_TAGS >> to expand tags, and
 provides fallbacks for the C<< :default >> and C<< :all >> tags.
 
+=item C<< _exporter_expand_regexp($regexp, $args, $globals) >>
+
+Like C<_exporter_expand_regexp>, but given a regexp-like string instead
+of a tag name.
+
+The default implementation greps through C<< @EXPORT_OK >>.
+
 =item C<< _exporter_expand_sub($name, $args, $globals) >>
 
 This method is called to translate a sub name to a hash of name => coderef
@@ -452,6 +572,39 @@ trying to handle all of it yourself.
 
 =back
 
+=head1 DIAGNOSTICS
+
+=over
+
+=item B<< Overwriting existing sub '%s::%s' with sub '%s' exported by %s >>
+
+A warning issued if Exporter::Tiny is asked to export a symbol which
+will result in an existing sub being overwritten. This warning can be
+suppressed using either of the following:
+
+   use MyUtils { replace => 1 }, "frobnicate";
+   use MyUtils "frobnicate" => { -replace => 1 };
+
+Or can be upgraded to a fatal error:
+
+   use MyUtils { replace => "die" }, "frobnicate";
+   use MyUtils "frobnicate" => { -replace => "die" };
+
+=item B<< Refusing to overwrite existing sub '%s::%s' with sub '%s' exported by %s >>
+
+The fatal version of the above warning.
+
+=item B<< Could not find sub '%s' exported by %s >>
+
+You requested to import a sub which the package does not provide.
+
+=item B<< Cannot provide an -as option for tags >>
+
+Because a tag may provide more than one function, it does not make sense
+to request a single name for it. Instead use C<< -prefix >> or C<< -suffix >>.
+
+=back
+
 =head1 HISTORY
 
 L<Type::Library> had a bunch of custom exporting code which poked coderefs
